@@ -16,11 +16,13 @@
  * The best way to set the system's RTC is so that it holds the current
  * time in UTC.  Use the "-l" flag to tell this program that the system
  * RTC uses a local timezone instead (maybe you dual-boot MS-Windows).
+ * That flag should not be needed on systems with adjtime support.
  */
 
 #include <stdio.h>
 #include <getopt.h>
 #include <fcntl.h>
+#include <libgen.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
@@ -46,10 +48,10 @@ static char		*progname;
 
 #define VERSION_STRING		"rtcwake from " PACKAGE_STRING
 #define RTC_PATH		"/sys/class/rtc/%s/device/power/wakeup"
-#define SYS_POWER_SATE_PATH	"/sys/power/state"
+#define SYS_POWER_STATE_PATH	"/sys/power/state"
 #define ADJTIME_PATH		"/etc/adjtime"
 #define DEFAULT_DEVICE		"/dev/rtc0"
-#define DEFAULT_MODE		"suspend"
+#define DEFAULT_MODE		"standby"
 
 enum ClockMode {
 	CM_AUTO,
@@ -89,7 +91,7 @@ static void usage(int retval)
 	exit(retval);
 }
 
-static int may_wakeup(const char *devname)
+static int is_wakeup_enabled(const char *devname)
 {
 	char	buf[128], *s;
 	FILE	*f;
@@ -136,12 +138,12 @@ static int get_basetimes(int fd)
 	 */
 	if (ioctl(fd, RTC_RD_TIME, &rtc) < 0) {
 		perror(_("read rtc time"));
-		return 0;
+		return -1;
 	}
 	sys_time = time(0);
 	if (sys_time == (time_t)-1) {
 		perror(_("read system time"));
-		return 0;
+		return -1;
 	}
 
 	/* convert rtc_time to normal arithmetic-friendly form,
@@ -159,22 +161,26 @@ static int get_basetimes(int fd)
 
 	if (rtc_time == (time_t)-1) {
 		perror(_("convert rtc time"));
-		return 0;
+		return -1;
 	}
 
 	if (verbose) {
-		if (clock_mode == CM_LOCAL) {
-			printf("\ttzone   = %ld\n", timezone);
-			printf("\ttzname  = %s\n", tzname[daylight]);
-			gmtime_r(&rtc_time, &tm);
-		}
-		printf("\tsystime = %ld, (UTC) %s\n",
+		/* Unless the system uses UTC, either delta or tzone
+		 * reflects a seconds offset from UTC.  The value can
+		 * help sort out problems like bugs in your C library.
+		 */
+		printf("\tdelta   = %ld\n", sys_time - rtc_time);
+		printf("\ttzone   = %ld\n", timezone);
+
+		printf("\ttzname  = %s\n", tzname[daylight]);
+		gmtime_r(&rtc_time, &tm);
+		printf("\tsystime = %ld, (UTC) %s",
 				(long) sys_time, asctime(gmtime(&sys_time)));
-		printf("\trtctime = %ld, (UTC) %s\n",
+		printf("\trtctime = %ld, (UTC) %s",
 				(long) rtc_time, asctime(&tm));
 	}
 
-	return 1;
+	return 0;
 }
 
 static int setup_alarm(int fd, time_t *wakeup)
@@ -182,7 +188,17 @@ static int setup_alarm(int fd, time_t *wakeup)
 	struct tm		*tm;
 	struct rtc_wkalrm	wake;
 
-	tm = gmtime(wakeup);
+	/* The wakeup time is in POSIX time (more or less UTC).
+	 * Ideally RTCs use that same time; but PCs can't do that
+	 * if they need to boot MS-Windows.  Messy...
+	 *
+	 * When clock_mode == CM_UTC this process's timezone is UTC,
+	 * so we'll pass a UTC date to the RTC.
+	 *
+	 * Else clock_mode == CM_LOCAL so the time given to the RTC
+	 * will instead use the local time zone.
+	 */
+	tm = localtime(wakeup);
 
 	wake.time.tm_sec = tm->tm_sec;
 	wake.time.tm_min = tm->tm_min;
@@ -190,41 +206,41 @@ static int setup_alarm(int fd, time_t *wakeup)
 	wake.time.tm_mday = tm->tm_mday;
 	wake.time.tm_mon = tm->tm_mon;
 	wake.time.tm_year = tm->tm_year;
-	wake.time.tm_wday = tm->tm_wday;
-	wake.time.tm_yday = tm->tm_yday;
-	wake.time.tm_isdst = tm->tm_isdst;
+	/* wday, yday, and isdst fields are unused by Linux */
+	wake.time.tm_wday = -1;
+	wake.time.tm_yday = -1;
+	wake.time.tm_isdst = -1;
 
-	/* many rtc alarms only support up to 24 hours from 'now' ... */
-	if ((rtc_time + (24 * 60 * 60)) > *wakeup) {
-		if (ioctl(fd, RTC_ALM_SET, &wake.time) < 0) {
-			perror(_("set rtc alarm"));
-			return 0;
-		}
-		if (ioctl(fd, RTC_AIE_ON, 0) < 0) {
-			perror(_("enable rtc alarm"));
-			return 0;
-		}
-
-		/* ... so use the "more than 24 hours" request only if we must */
-	} else {
-		/* avoid an extra AIE_ON call */
-		wake.enabled = 1;
-
-		if (ioctl(fd, RTC_WKALM_SET, &wake) < 0) {
+	wake.enabled = 1;
+	/* First try the preferred RTC_WKALM_SET */
+	if (ioctl(fd, RTC_WKALM_SET, &wake) < 0) {
+		wake.enabled = 0;
+		/* Fall back on the non-preferred way of setting wakeups; only
+		* works for alarms < 24 hours from now */
+		if ((rtc_time + (24 * 60 * 60)) > *wakeup) {
+			if (ioctl(fd, RTC_ALM_SET, &wake.time) < 0) {
+				perror(_("set rtc alarm"));
+				return -1;
+			}
+			if (ioctl(fd, RTC_AIE_ON, 0) < 0) {
+				perror(_("enable rtc alarm"));
+				return -1;
+			}
+		} else {
 			perror(_("set rtc wake alarm"));
-			return 0;
+			return -1;
 		}
 	}
 
-	return 1;
+	return 0;
 }
 
 static void suspend_system(const char *suspend)
 {
-	FILE	*f = fopen(SYS_POWER_SATE_PATH, "w");
+	FILE	*f = fopen(SYS_POWER_STATE_PATH, "w");
 
 	if (!f) {
-		perror(SYS_POWER_SATE_PATH);
+		perror(SYS_POWER_STATE_PATH);
 		return;
 	}
 
@@ -243,24 +259,24 @@ static int read_clock_mode(void)
 
 	fp = fopen(ADJTIME_PATH, "r");
 	if (!fp)
-		return 0;
+		return -1;
 
 	/* skip first line */
 	if (!fgets(linebuf, MAX_LINE, fp)) {
 		fclose(fp);
-		return 0;
+		return -1;
 	}
 
 	/* skip second line */
 	if (!fgets(linebuf, MAX_LINE, fp)) {
 		fclose(fp);
-		return 0;
+		return -1;
 	}
 
 	/* read third line */
 	if (!fgets(linebuf, MAX_LINE, fp)) {
 		fclose(fp);
-		return 0;
+		return -1;
 	}
 
 	if (strncmp(linebuf, "UTC", 3) == 0)
@@ -270,7 +286,7 @@ static int read_clock_mode(void)
 
 	fclose(fp);
 
-	return 1;
+	return 0;
 }
 
 int main(int argc, char **argv)
@@ -317,12 +333,14 @@ int main(int argc, char **argv)
 					|| strcmp(optarg, "mem") == 0
 					|| strcmp(optarg, "disk") == 0
 					|| strcmp(optarg, "on") == 0
+					|| strcmp(optarg, "no") == 0
 			   ) {
 				suspend = strdup(optarg);
 				break;
 			}
-			fprintf(stderr, _("%s: unrecognized suspend state '%s'\n"),
-					progname, optarg);
+			fprintf(stderr,
+				_("%s: unrecognized suspend state '%s'\n"),
+				progname, optarg);
 			usage(EXIT_FAILURE);
 
 			/* alarm time, seconds-to-sleep (relative) */
@@ -372,14 +390,14 @@ int main(int argc, char **argv)
 	}
 
 	if (clock_mode == CM_AUTO) {
-		if (!read_clock_mode()) {
+		if (read_clock_mode() < 0) {
 			printf(_("%s: assuming RTC uses UTC ...\n"), progname);
 			clock_mode = CM_UTC;
 		}
-		if (verbose)
-			printf(_("Using %s time\n"),
-					clock_mode == CM_UTC ? "UTC" : _("local"));
 	}
+	if (verbose)
+		printf(clock_mode == CM_UTC ? _("Using UTC time.\n") :
+				_("Using local time.\n"));
 
 	if (!alarm && !seconds) {
 		fprintf(stderr, _("%s: must provide wake time\n"), progname);
@@ -402,7 +420,8 @@ int main(int argc, char **argv)
 		devname = new_devname;
 	}
 
-	if (strcmp(suspend, "on") != 0 && !may_wakeup(devname)) {
+	if (strcmp(suspend, "on") != 0 && strcmp(suspend, "no") != 0
+			&& !is_wakeup_enabled(devname)) {
 		fprintf(stderr, _("%s: %s not enabled for wakeup events\n"),
 				progname, devname);
 		exit(EXIT_FAILURE);
@@ -416,15 +435,16 @@ int main(int argc, char **argv)
 	}
 
 	/* relative or absolute alarm time, normalized to time_t */
-	if (!get_basetimes(fd))
+	if (get_basetimes(fd) < 0)
 		exit(EXIT_FAILURE);
 	if (verbose)
 		printf(_("alarm %ld, sys_time %ld, rtc_time %ld, seconds %u\n"),
 				alarm, sys_time, rtc_time, seconds);
 	if (alarm) {
 		if (alarm < sys_time) {
-			fprintf(stderr, _("%s: time doesn't go backward to %s\n"),
-					progname, ctime(&alarm));
+			fprintf(stderr,
+				_("%s: time doesn't go backward to %s\n"),
+				progname, ctime(&alarm));
 			exit(EXIT_FAILURE);
 		}
 		alarm += sys_time - rtc_time;
@@ -433,16 +453,18 @@ int main(int argc, char **argv)
 	if (setup_alarm(fd, &alarm) < 0)
 		exit(EXIT_FAILURE);
 
-	sync();
 	printf(_("%s: wakeup from \"%s\" using %s at %s\n"),
 			progname, suspend, devname,
 			ctime(&alarm));
 	fflush(stdout);
 	usleep(10 * 1000);
 
-	if (strcmp(suspend, "on") != 0)
+	if (strcmp(suspend, "no") == 0)
+		exit(EXIT_SUCCESS);
+	else if (strcmp(suspend, "on") != 0) {
+		sync();
 		suspend_system(suspend);
-	else {
+	} else {
 		unsigned long data;
 
 		do {
